/*
 * SPDX-FileCopyrightText: (c) 2003, 2004 Lev Walkin <vlm@lionet.info>. All rights reserved.
 * SPDX-License-Identifier: BSD-1-Clause
 * Generated by asn1c-0.9.22 (http://lionet.info/asn1c)
 * From ASN.1 module "ULP-Components"
 *     found in "../supl-common.asn"
 */

#include "IPAddress.h"

static int memb_ipv4Address_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    const OCTET_STRING_t *st = (const OCTET_STRING_t *)sptr;
    size_t size;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    size = st->size;

    if (size == 4)
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ipv6Address_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    const OCTET_STRING_t *st = (const OCTET_STRING_t *)sptr;
    size_t size;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    size = st->size;

    if (size == 16)
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static asn_per_constraints_t ASN_PER_MEMB_IPV4_ADDRESS_CONSTR_2 = {
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    {APC_CONSTRAINED, 0, 0, 4, 4} /* (SIZE(4..4)) */,
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_IPV6_ADDRESS_CONSTR_3 = {
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    {APC_CONSTRAINED, 0, 0, 16, 16} /* (SIZE(16..16)) */,
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_TYPE_IP_ADDRESS_CONSTR_1 = {
    {APC_CONSTRAINED, 1, 1, 0, 1} /* (0..1) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_TYPE_member_t asn_MBR_IPAddress_1[] = {
    {ATF_NOFLAGS, 0, offsetof(struct IPAddress, choice.ipv4Address),
        (ASN_TAG_CLASS_CONTEXT | (0 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_OCTET_STRING, memb_ipv4Address_constraint_1,
        &ASN_PER_MEMB_IPV4_ADDRESS_CONSTR_2, 0, "ipv4Address"},
    {ATF_NOFLAGS, 0, offsetof(struct IPAddress, choice.ipv6Address),
        (ASN_TAG_CLASS_CONTEXT | (1 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_OCTET_STRING, memb_ipv6Address_constraint_1,
        &ASN_PER_MEMB_IPV6_ADDRESS_CONSTR_3, 0, "ipv6Address"},
};
static asn_TYPE_tag2member_t asn_MAP_IPAddress_tag2el_1[] = {
    {(ASN_TAG_CLASS_CONTEXT | (0 << 2)), 0, 0, 0}, /* ipv4Address at 41 */
    {(ASN_TAG_CLASS_CONTEXT | (1 << 2)), 1, 0, 0}  /* ipv6Address at 42 */
};
static asn_CHOICE_specifics_t asn_SPC_IPAddress_specs_1 = {
    sizeof(struct IPAddress),
    offsetof(struct IPAddress, _asn_ctx),
    offsetof(struct IPAddress, present),
    sizeof(((struct IPAddress *)0)->present),
    asn_MAP_IPAddress_tag2el_1,
    2, /* Count of tags in the map */
    0,
    -1 /* Extensions start */
};
asn_TYPE_descriptor_t asn_DEF_IPAddress = {
    "IPAddress",
    "IPAddress",
    CHOICE_free,
    CHOICE_print,
    CHOICE_constraint,
    CHOICE_decode_ber,
    CHOICE_encode_der,
    CHOICE_decode_xer,
    CHOICE_encode_xer,
    CHOICE_decode_uper,
    CHOICE_encode_uper,
    CHOICE_outmost_tag,
    0, /* No effective tags (pointer) */
    0, /* No effective tags (count) */
    0, /* No tags (pointer) */
    0, /* No tags (count) */
    &ASN_PER_TYPE_IP_ADDRESS_CONSTR_1,
    asn_MBR_IPAddress_1,
    2,                         /* Elements count */
    &asn_SPC_IPAddress_specs_1 /* Additional specs */
};
