/*
 * SPDX-FileCopyrightText: (c) 2003, 2004 Lev Walkin <vlm@lionet.info>. All rights reserved.
 * SPDX-License-Identifier: BSD-1-Clause
 * Generated by asn1c-0.9.22 (http://lionet.info/asn1c)
 * From ASN.1 module "RRLP-Components"
 *     found in "../rrlp-components.asn"
 */

#ifndef _ControlHeader_H
#define _ControlHeader_H

#include <asn_application.h>

/* Including external dependencies */
#include <constr_SEQUENCE.h>

#ifdef __cplusplus
extern "C"
{
#endif

    /* Forward declarations */
    struct ReferenceTime;
    struct RefLocation;
    struct DGPSCorrections;
    struct NavigationModel;
    struct IonosphericModel;
    struct UTCModel;
    struct Almanac;
    struct AcquisAssist;
    struct SeqOf_BadSatelliteSet;

    /* ControlHeader */
    typedef struct ControlHeader
    {
        struct ReferenceTime *referenceTime /* OPTIONAL */;
        struct RefLocation *refLocation /* OPTIONAL */;
        struct DGPSCorrections *dgpsCorrections /* OPTIONAL */;
        struct NavigationModel *navigationModel /* OPTIONAL */;
        struct IonosphericModel *ionosphericModel /* OPTIONAL */;
        struct UTCModel *utcModel /* OPTIONAL */;
        struct Almanac *almanac /* OPTIONAL */;
        struct AcquisAssist *acquisAssist /* OPTIONAL */;
        struct SeqOf_BadSatelliteSet *realTimeIntegrity /* OPTIONAL */;

        /* Context for parsing across buffer boundaries */
        asn_struct_ctx_t _asn_ctx;
    } ControlHeader_t;

    /* Implementation */
    extern asn_TYPE_descriptor_t asn_DEF_ControlHeader;

#ifdef __cplusplus
}
#endif

/* Referred external types */
#include "AcquisAssist.h"
#include "Almanac.h"
#include "DGPSCorrections.h"
#include "IonosphericModel.h"
#include "NavigationModel.h"
#include "RefLocation.h"
#include "ReferenceTime.h"
#include "SeqOf-BadSatelliteSet.h"
#include "UTCModel.h"

#endif /* _ControlHeader_H_ */
#include <asn_internal.h>
