/*
 * SPDX-FileCopyrightText: (c) 2003, 2004 Lev Walkin <vlm@lionet.info>. All rights reserved.
 * SPDX-License-Identifier: BSD-1-Clause
 * Generated by asn1c-0.9.22 (http://lionet.info/asn1c)
 * From ASN.1 module "RRLP-Components"
 *     found in "../rrlp-components.asn"
 */

#ifndef _Ext_GeographicalInformation_H
#define _Ext_GeographicalInformation_H

#include <asn_application.h>

/* Including external dependencies */
#include <OCTET_STRING.h>

#ifdef __cplusplus
extern "C"
{
#endif

    /* Ext-GeographicalInformation */
    typedef OCTET_STRING_t Ext_GeographicalInformation_t;

    /* Implementation */
    extern asn_TYPE_descriptor_t asn_DEF_Ext_GeographicalInformation;
    asn_struct_free_f Ext_GeographicalInformation_free;
    asn_struct_print_f Ext_GeographicalInformation_print;
    asn_constr_check_f Ext_GeographicalInformation_constraint;
    ber_type_decoder_f Ext_GeographicalInformation_decode_ber;
    der_type_encoder_f Ext_GeographicalInformation_encode_der;
    xer_type_decoder_f Ext_GeographicalInformation_decode_xer;
    xer_type_encoder_f Ext_GeographicalInformation_encode_xer;
    per_type_decoder_f Ext_GeographicalInformation_decode_uper;
    per_type_encoder_f Ext_GeographicalInformation_encode_uper;

#ifdef __cplusplus
}
#endif

#endif /* _Ext_GeographicalInformation_H_ */
#include <asn_internal.h>
