/*
 * SPDX-FileCopyrightText: (c) 2003, 2004 Lev Walkin <vlm@lionet.info>. All rights reserved.
 * SPDX-License-Identifier: BSD-1-Clause
 * Generated by asn1c-0.9.22 (http://lionet.info/asn1c)
 * From ASN.1 module "RRLP-Components"
 *     found in "../rrlp-components.asn"
 */

#include "GANSS-SgnElement.h"

static int memb_cNo_constraint_1(asn_TYPE_descriptor_t *td, const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= 0 && value <= 63))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_carrierQualityInd_constraint_1(
    asn_TYPE_descriptor_t *td, const void *sptr,
    asn_app_constraint_failed_f *ctfailcb, void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= 0 && value <= 3))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_codePhase_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= 0 && value <= 2097151))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_integerCodePhase_constraint_1(
    asn_TYPE_descriptor_t *td, const void *sptr,
    asn_app_constraint_failed_f *ctfailcb, void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= 0 && value <= 63))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_codePhaseRMSError_constraint_1(
    asn_TYPE_descriptor_t *td, const void *sptr,
    asn_app_constraint_failed_f *ctfailcb, void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= 0 && value <= 63))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_doppler_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= -32768 && value <= 32767))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_adr_constraint_1(asn_TYPE_descriptor_t *td, const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= 0 && value <= 33554431))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static asn_per_constraints_t ASN_PER_MEMB_C_NO_CONSTR_3 = {
    {APC_CONSTRAINED, 6, 6, 0, 63} /* (0..63) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_CARRIER_QUALITY_IND_CONSTR_5 = {
    {APC_CONSTRAINED, 2, 2, 0, 3} /* (0..3) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_CODE_PHASE_CONSTR_6 = {
    {APC_CONSTRAINED, 21, -1, 0, 2097151} /* (0..2097151) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_INTEGER_CODE_PHASE_CONSTR_7 = {
    {APC_CONSTRAINED, 6, 6, 0, 63} /* (0..63) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_CODE_PHASE_RMS_ERROR_CONSTR_8 = {
    {APC_CONSTRAINED, 6, 6, 0, 63} /* (0..63) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_DOPPLER_CONSTR_9 = {
    {APC_CONSTRAINED, 16, 16, -32768, 32767} /* (-32768..32767) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_ADR_CONSTR_10 = {
    {APC_CONSTRAINED, 25, -1, 0, 33554431} /* (0..33554431) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_TYPE_member_t asn_MBR_GANSS_SgnElement_1[] = {
    {ATF_NOFLAGS, 0, offsetof(struct GANSS_SgnElement, svID),
        (ASN_TAG_CLASS_CONTEXT | (0 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_SVID, 0,                       /* Defer constraints checking to the member type */
        0,                                      /* No PER visible constraints */
        0, "svID"},
    {ATF_NOFLAGS, 0, offsetof(struct GANSS_SgnElement, cNo),
        (ASN_TAG_CLASS_CONTEXT | (1 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_cNo_constraint_1, &ASN_PER_MEMB_C_NO_CONSTR_3,
        0, "cNo"},
    {ATF_NOFLAGS, 0, offsetof(struct GANSS_SgnElement, mpathDet),
        (ASN_TAG_CLASS_CONTEXT | (2 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_MpathIndic, 0,                 /* Defer constraints checking to the member type */
        0,                                      /* No PER visible constraints */
        0, "mpathDet"},
    {ATF_POINTER, 1, offsetof(struct GANSS_SgnElement, carrierQualityInd),
        (ASN_TAG_CLASS_CONTEXT | (3 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_carrierQualityInd_constraint_1,
        &ASN_PER_MEMB_CARRIER_QUALITY_IND_CONSTR_5, 0, "carrierQualityInd"},
    {ATF_NOFLAGS, 0, offsetof(struct GANSS_SgnElement, codePhase),
        (ASN_TAG_CLASS_CONTEXT | (4 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_codePhase_constraint_1,
        &ASN_PER_MEMB_CODE_PHASE_CONSTR_6, 0, "codePhase"},
    {ATF_POINTER, 1, offsetof(struct GANSS_SgnElement, integerCodePhase),
        (ASN_TAG_CLASS_CONTEXT | (5 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_integerCodePhase_constraint_1,
        &ASN_PER_MEMB_INTEGER_CODE_PHASE_CONSTR_7, 0, "integerCodePhase"},
    {ATF_NOFLAGS, 0, offsetof(struct GANSS_SgnElement, codePhaseRMSError),
        (ASN_TAG_CLASS_CONTEXT | (6 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_codePhaseRMSError_constraint_1,
        &ASN_PER_MEMB_CODE_PHASE_RMS_ERROR_CONSTR_8, 0, "codePhaseRMSError"},
    {ATF_POINTER, 2, offsetof(struct GANSS_SgnElement, doppler),
        (ASN_TAG_CLASS_CONTEXT | (7 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_doppler_constraint_1,
        &ASN_PER_MEMB_DOPPLER_CONSTR_9, 0, "doppler"},
    {ATF_POINTER, 1, offsetof(struct GANSS_SgnElement, adr),
        (ASN_TAG_CLASS_CONTEXT | (8 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_adr_constraint_1, &ASN_PER_MEMB_ADR_CONSTR_10,
        0, "adr"},
};
static int asn_MAP_GANSS_SgnElement_oms_1[] = {3, 5, 7, 8};
static ber_tlv_tag_t asn_DEF_GANSS_SgnElement_tags_1[] = {
    (ASN_TAG_CLASS_UNIVERSAL | (16 << 2))};
static asn_TYPE_tag2member_t asn_MAP_GANSS_SgnElement_tag2el_1[] = {
    {(ASN_TAG_CLASS_CONTEXT | (0 << 2)), 0, 0, 0}, /* svID at 1479 */
    {(ASN_TAG_CLASS_CONTEXT | (1 << 2)), 1, 0, 0}, /* cNo at 1480 */
    {(ASN_TAG_CLASS_CONTEXT | (2 << 2)), 2, 0, 0}, /* mpathDet at 1481 */
    {(ASN_TAG_CLASS_CONTEXT | (3 << 2)), 3, 0,
        0},                                        /* carrierQualityInd at 1482 */
    {(ASN_TAG_CLASS_CONTEXT | (4 << 2)), 4, 0, 0}, /* codePhase at 1483 */
    {(ASN_TAG_CLASS_CONTEXT | (5 << 2)), 5, 0,
        0}, /* integerCodePhase at 1484 */
    {(ASN_TAG_CLASS_CONTEXT | (6 << 2)), 6, 0,
        0},                                        /* codePhaseRMSError at 1485 */
    {(ASN_TAG_CLASS_CONTEXT | (7 << 2)), 7, 0, 0}, /* doppler at 1486 */
    {(ASN_TAG_CLASS_CONTEXT | (8 << 2)), 8, 0, 0}  /* adr at 1487 */
};
static asn_SEQUENCE_specifics_t asn_SPC_GANSS_SgnElement_specs_1 = {
    sizeof(struct GANSS_SgnElement),
    offsetof(struct GANSS_SgnElement, _asn_ctx),
    asn_MAP_GANSS_SgnElement_tag2el_1,
    9,                              /* Count of tags in the map */
    asn_MAP_GANSS_SgnElement_oms_1, /* Optional members */
    4,
    0,  /* Root/Additions */
    -1, /* Start extensions */
    -1  /* Stop extensions */
};
asn_TYPE_descriptor_t asn_DEF_GANSS_SgnElement = {
    "GANSS-SgnElement",
    "GANSS-SgnElement",
    SEQUENCE_free,
    SEQUENCE_print,
    SEQUENCE_constraint,
    SEQUENCE_decode_ber,
    SEQUENCE_encode_der,
    SEQUENCE_decode_xer,
    SEQUENCE_encode_xer,
    SEQUENCE_decode_uper,
    SEQUENCE_encode_uper,
    0, /* Use generic outmost tag fetcher */
    asn_DEF_GANSS_SgnElement_tags_1,
    sizeof(asn_DEF_GANSS_SgnElement_tags_1) /
        sizeof(asn_DEF_GANSS_SgnElement_tags_1[0]), /* 1 */
    asn_DEF_GANSS_SgnElement_tags_1,                /* Same as above */
    sizeof(asn_DEF_GANSS_SgnElement_tags_1) /
        sizeof(asn_DEF_GANSS_SgnElement_tags_1[0]), /* 1 */
    0,                                              /* No PER visible constraints */
    asn_MBR_GANSS_SgnElement_1,
    9,                                /* Elements count */
    &asn_SPC_GANSS_SgnElement_specs_1 /* Additional specs */
};
