/*
 * SPDX-FileCopyrightText: (c) 2003, 2004 Lev Walkin <vlm@lionet.info>. All rights reserved.
 * SPDX-License-Identifier: BSD-1-Clause
 * Generated by asn1c-0.9.22 (http://lionet.info/asn1c)
 * From ASN.1 module "SUPL-START"
 *     found in "../supl-start.asn"
 */

#ifndef _SUPLSTART_H
#define _SUPLSTART_H

#include <asn_application.h>

/* Including external dependencies */
#include "LocationId.h"
#include "SETCapabilities.h"
#include <constr_SEQUENCE.h>

#ifdef __cplusplus
extern "C"
{
#endif

    /* Forward declarations */
    struct QoP;

    /* SUPLSTART */
    typedef struct SUPLSTART
    {
        SETCapabilities_t sETCapabilities;
        LocationId_t locationId;
        struct QoP *qoP /* OPTIONAL */;
        /*
         * This type is extensible,
         * possible extensions are below.
         */

        /* Context for parsing across buffer boundaries */
        asn_struct_ctx_t _asn_ctx;
    } SUPLSTART_t;

    /* Implementation */
    extern asn_TYPE_descriptor_t asn_DEF_SUPLSTART;

#ifdef __cplusplus
}
#endif

/* Referred external types */
#include "QoP.h"

#endif /* _SUPLSTART_H_ */
#include <asn_internal.h>
