/*
 * SPDX-FileCopyrightText: (c) 2003, 2004 Lev Walkin <vlm@lionet.info>. All rights reserved.
 * SPDX-License-Identifier: BSD-1-Clause
 * Generated by asn1c-0.9.22 (http://lionet.info/asn1c)
 * From ASN.1 module "RRLP-Components"
 *     found in "../rrlp-components.asn"
 */

#include "MethodType.h"

static asn_per_constraints_t ASN_PER_TYPE_METHOD_TYPE_CONSTR_1 = {
    {APC_CONSTRAINED, 2, 2, 0, 3} /* (0..3) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_TYPE_member_t asn_MBR_MethodType_1[] = {
    {ATF_NOFLAGS, 0, offsetof(struct MethodType, choice.msAssisted),
        (ASN_TAG_CLASS_CONTEXT | (0 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_AccuracyOpt,
        0, /* Defer constraints checking to the member type */
        0, /* No PER visible constraints */
        0, "msAssisted"},
    {ATF_NOFLAGS, 0, offsetof(struct MethodType, choice.msBased),
        (ASN_TAG_CLASS_CONTEXT | (1 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_Accuracy, 0,                   /* Defer constraints checking to the member type */
        0,                                      /* No PER visible constraints */
        0, "msBased"},
    {ATF_NOFLAGS, 0, offsetof(struct MethodType, choice.msBasedPref),
        (ASN_TAG_CLASS_CONTEXT | (2 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_Accuracy, 0,                   /* Defer constraints checking to the member type */
        0,                                      /* No PER visible constraints */
        0, "msBasedPref"},
    {ATF_NOFLAGS, 0, offsetof(struct MethodType, choice.msAssistedPref),
        (ASN_TAG_CLASS_CONTEXT | (3 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_Accuracy, 0,                   /* Defer constraints checking to the member type */
        0,                                      /* No PER visible constraints */
        0, "msAssistedPref"},
};
static asn_TYPE_tag2member_t asn_MAP_MethodType_tag2el_1[] = {
    {(ASN_TAG_CLASS_CONTEXT | (0 << 2)), 0, 0, 0}, /* msAssisted at 117 */
    {(ASN_TAG_CLASS_CONTEXT | (1 << 2)), 1, 0, 0}, /* msBased at 118 */
    {(ASN_TAG_CLASS_CONTEXT | (2 << 2)), 2, 0, 0}, /* msBasedPref at 119 */
    {(ASN_TAG_CLASS_CONTEXT | (3 << 2)), 3, 0, 0}  /* msAssistedPref at 121 */
};
static asn_CHOICE_specifics_t asn_SPC_MethodType_specs_1 = {
    sizeof(struct MethodType),
    offsetof(struct MethodType, _asn_ctx),
    offsetof(struct MethodType, present),
    sizeof(((struct MethodType *)0)->present),
    asn_MAP_MethodType_tag2el_1,
    4, /* Count of tags in the map */
    0,
    -1 /* Extensions start */
};
asn_TYPE_descriptor_t asn_DEF_MethodType = {
    "MethodType",
    "MethodType",
    CHOICE_free,
    CHOICE_print,
    CHOICE_constraint,
    CHOICE_decode_ber,
    CHOICE_encode_der,
    CHOICE_decode_xer,
    CHOICE_encode_xer,
    CHOICE_decode_uper,
    CHOICE_encode_uper,
    CHOICE_outmost_tag,
    0, /* No effective tags (pointer) */
    0, /* No effective tags (count) */
    0, /* No tags (pointer) */
    0, /* No tags (count) */
    &ASN_PER_TYPE_METHOD_TYPE_CONSTR_1,
    asn_MBR_MethodType_1,
    4,                          /* Elements count */
    &asn_SPC_MethodType_specs_1 /* Additional specs */
};
