/*
 * SPDX-FileCopyrightText: (c) 2003, 2004 Lev Walkin <vlm@lionet.info>. All rights reserved.
 * SPDX-License-Identifier: BSD-1-Clause
 * Generated by asn1c-0.9.22 (http://lionet.info/asn1c)
 * From ASN.1 module "SUPL-INIT"
 *     found in "../supl-init.asn"
 */

#ifndef _Notification_H
#define _Notification_H

#include <asn_application.h>

/* Including external dependencies */
#include "EncodingType.h"
#include "FormatIndicator.h"
#include "NotificationType.h"
#include <OCTET_STRING.h>
#include <constr_SEQUENCE.h>

#ifdef __cplusplus
extern "C"
{
#endif

    /* Notification */
    typedef struct Notification
    {
        NotificationType_t notificationType;
        EncodingType_t *encodingType /* OPTIONAL */;
        OCTET_STRING_t *requestorId /* OPTIONAL */;
        FormatIndicator_t *requestorIdType /* OPTIONAL */;
        OCTET_STRING_t *clientName /* OPTIONAL */;
        FormatIndicator_t *clientNameType /* OPTIONAL */;
        /*
         * This type is extensible,
         * possible extensions are below.
         */

        /* Context for parsing across buffer boundaries */
        asn_struct_ctx_t _asn_ctx;
    } Notification_t;

    /* Implementation */
    extern asn_TYPE_descriptor_t asn_DEF_Notification;

#ifdef __cplusplus
}
#endif

#endif /* _Notification_H_ */
#include <asn_internal.h>
