/*-
 * SPDX-FileCopyrightText: (c) 2003, 2004 Lev Walkin <vlm@lionet.info>. All rights reserved.
 * SPDX-License-Identifier: BSD-1-Clause
 */
#ifndef _GeneralizedTime_H
#define _GeneralizedTime_H

#include <OCTET_STRING.h>

#ifdef __cplusplus
extern "C"
{
#endif

    typedef OCTET_STRING_t GeneralizedTime_t; /* Implemented via OCTET STRING */

    extern asn_TYPE_descriptor_t asn_DEF_GeneralizedTime;

    asn_struct_print_f GeneralizedTime_print;
    asn_constr_check_f GeneralizedTime_constraint;
    der_type_encoder_f GeneralizedTime_encode_der;
    xer_type_encoder_f GeneralizedTime_encode_xer;

    /***********************
     * Some handy helpers. *
     ***********************/

    struct tm; /* <time.h> */

    /*
     * Convert a GeneralizedTime structure into time_t
     * and optionally into struct tm.
     * If as_gmt is given, the resulting _optional_tm4fill will have a GMT zone,
     * instead of default local one.
     * On error returns -1 and errno set to EINVAL
     */
    time_t asn_GT2time(const GeneralizedTime_t * /*st*/, struct tm *ret_tm,
        int as_gmt);

    /* A version of the above function also returning the fractions of seconds
     */
    time_t asn_GT2time_frac(const GeneralizedTime_t * /*st*/, int *frac_value,
        int *frac_digits, /* (value / (10 ^ digits)) */
        struct tm *ret_tm, int as_gmt);

    /*
     * Another version returning fractions with defined precision
     * For example, parsing of the time ending with ".1" seconds
     * with frac_digits=3 (msec) would yield frac_value = 100.
     */
    time_t asn_GT2time_prec(const GeneralizedTime_t * /*st*/, int *frac_value,
        int frac_digits, struct tm *ret_tm, int as_gmt);

    /*
     * Convert a struct tm into GeneralizedTime.
     * If _optional_gt is not given, this function will try to allocate one.
     * If force_gmt is given, the resulting GeneralizedTime will be forced
     * into a GMT time zone (encoding ends with a "Z").
     * On error, this function returns 0 and sets errno.
     */
    GeneralizedTime_t *asn_time2GT(GeneralizedTime_t *opt_gt,
        const struct tm * /*tm*/, int force_gmt);
    GeneralizedTime_t *asn_time2GT_frac(GeneralizedTime_t *opt_gt,
        const struct tm * /*tm*/,
        int frac_value, int frac_digits,
        int force_gmt);

#ifdef __cplusplus
}
#endif

#endif /* _GeneralizedTime_H_ */
