/*
 * SPDX-FileCopyrightText: (c) 2003, 2004 Lev Walkin <vlm@lionet.info>. All rights reserved.
 * SPDX-License-Identifier: BSD-1-Clause
 * Generated by asn1c-0.9.22 (http://lionet.info/asn1c)
 * From ASN.1 module "RRLP-Components"
 *     found in "../rrlp-components.asn"
 */

#include "NavModel-KeplerianSet.h"

static int memb_keplerToeLSB_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= 0 && value <= 511))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_keplerW_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= (-2147483647L - 1) && value <= 2147483647))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_keplerDeltaN_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= -32768 && value <= 32767))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_keplerM0_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= (-2147483647L - 1) && value <= 2147483647))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_keplerOmegaDot_constraint_1(
    asn_TYPE_descriptor_t *td, const void *sptr,
    asn_app_constraint_failed_f *ctfailcb, void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= -8388608 && value <= 8388607))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_keplerELSB_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= 0 && value <= 33554431))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_keplerIDot_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= -8192 && value <= 8191))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_keplerAPowerHalfLSB_constraint_1(
    asn_TYPE_descriptor_t *td, const void *sptr,
    asn_app_constraint_failed_f *ctfailcb, void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= 0 && value <= 67108863))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_keplerI0_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= (-2147483647L - 1) && value <= 2147483647))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_keplerOmega0_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= (-2147483647L - 1) && value <= 2147483647))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_keplerCrs_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= -32768 && value <= 32767))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_keplerCis_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= -32768 && value <= 32767))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_keplerCus_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= -32768 && value <= 32767))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_keplerCrc_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= -32768 && value <= 32767))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_keplerCic_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= -32768 && value <= 32767))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_keplerCuc_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= -32768 && value <= 32767))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static asn_per_constraints_t ASN_PER_MEMB_KEPLER_TOE_LSB_CONSTR_2 = {
    {APC_CONSTRAINED, 9, 9, 0, 511} /* (0..511) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_KEPLER_W_CONSTR_3 = {
    {APC_CONSTRAINED, 32, -1, (-2147483647L - 1),
        2147483647} /* (-2147483648..2147483647) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_KEPLER_DELTA_N_CONSTR_4 = {
    {APC_CONSTRAINED, 16, 16, -32768, 32767} /* (-32768..32767) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_KEPLER_M0_CONSTR_5 = {
    {APC_CONSTRAINED, 32, -1, (-2147483647L - 1),
        2147483647} /* (-2147483648..2147483647) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_KEPLER_OMEGA_DOT_CONSTR_6 = {
    {APC_CONSTRAINED, 24, -1, -8388608, 8388607} /* (-8388608..8388607) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_KEPLER_ELSB_CONSTR_7 = {
    {APC_CONSTRAINED, 25, -1, 0, 33554431} /* (0..33554431) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_KEPLER_I_DOT_CONSTR_8 = {
    {APC_CONSTRAINED, 14, 14, -8192, 8191} /* (-8192..8191) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_KEPLER_A_POWER_HALF_LSB_CONSTR_9 = {
    {APC_CONSTRAINED, 26, -1, 0, 67108863} /* (0..67108863) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_KEPLER_I0_CONSTR_10 = {
    {APC_CONSTRAINED, 32, -1, (-2147483647L - 1),
        2147483647} /* (-2147483648..2147483647) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_KEPLER_OMEGA0_CONSTR_11 = {
    {APC_CONSTRAINED, 32, -1, (-2147483647L - 1),
        2147483647} /* (-2147483648..2147483647) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_KEPLER_CRS_CONSTR_12 = {
    {APC_CONSTRAINED, 16, 16, -32768, 32767} /* (-32768..32767) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_KEPLER_CIS_CONSTR_13 = {
    {APC_CONSTRAINED, 16, 16, -32768, 32767} /* (-32768..32767) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_KEPLER_CUS_CONSTR_14 = {
    {APC_CONSTRAINED, 16, 16, -32768, 32767} /* (-32768..32767) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_KEPLER_CRC_CONSTR_15 = {
    {APC_CONSTRAINED, 16, 16, -32768, 32767} /* (-32768..32767) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_KEPLER_CIC_CONSTR_16 = {
    {APC_CONSTRAINED, 16, 16, -32768, 32767} /* (-32768..32767) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_KEPLER_CUC_CONSTR_17 = {
    {APC_CONSTRAINED, 16, 16, -32768, 32767} /* (-32768..32767) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_TYPE_member_t asn_MBR_NavModel_KeplerianSet_1[] = {
    {ATF_NOFLAGS, 0, offsetof(struct NavModel_KeplerianSet, keplerToeLSB),
        (ASN_TAG_CLASS_CONTEXT | (0 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_keplerToeLSB_constraint_1,
        &ASN_PER_MEMB_KEPLER_TOE_LSB_CONSTR_2, 0, "keplerToeLSB"},
    {ATF_NOFLAGS, 0, offsetof(struct NavModel_KeplerianSet, keplerW),
        (ASN_TAG_CLASS_CONTEXT | (1 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_keplerW_constraint_1,
        &ASN_PER_MEMB_KEPLER_W_CONSTR_3, 0, "keplerW"},
    {ATF_NOFLAGS, 0, offsetof(struct NavModel_KeplerianSet, keplerDeltaN),
        (ASN_TAG_CLASS_CONTEXT | (2 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_keplerDeltaN_constraint_1,
        &ASN_PER_MEMB_KEPLER_DELTA_N_CONSTR_4, 0, "keplerDeltaN"},
    {ATF_NOFLAGS, 0, offsetof(struct NavModel_KeplerianSet, keplerM0),
        (ASN_TAG_CLASS_CONTEXT | (3 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_keplerM0_constraint_1,
        &ASN_PER_MEMB_KEPLER_M0_CONSTR_5, 0, "keplerM0"},
    {ATF_NOFLAGS, 0, offsetof(struct NavModel_KeplerianSet, keplerOmegaDot),
        (ASN_TAG_CLASS_CONTEXT | (4 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_keplerOmegaDot_constraint_1,
        &ASN_PER_MEMB_KEPLER_OMEGA_DOT_CONSTR_6, 0, "keplerOmegaDot"},
    {ATF_NOFLAGS, 0, offsetof(struct NavModel_KeplerianSet, keplerELSB),
        (ASN_TAG_CLASS_CONTEXT | (5 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_keplerELSB_constraint_1,
        &ASN_PER_MEMB_KEPLER_ELSB_CONSTR_7, 0, "keplerELSB"},
    {ATF_NOFLAGS, 0, offsetof(struct NavModel_KeplerianSet, keplerIDot),
        (ASN_TAG_CLASS_CONTEXT | (6 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_keplerIDot_constraint_1,
        &ASN_PER_MEMB_KEPLER_I_DOT_CONSTR_8, 0, "keplerIDot"},
    {ATF_NOFLAGS, 0,
        offsetof(struct NavModel_KeplerianSet, keplerAPowerHalfLSB),
        (ASN_TAG_CLASS_CONTEXT | (7 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_keplerAPowerHalfLSB_constraint_1,
        &ASN_PER_MEMB_KEPLER_A_POWER_HALF_LSB_CONSTR_9, 0, "keplerAPowerHalfLSB"},
    {ATF_NOFLAGS, 0, offsetof(struct NavModel_KeplerianSet, keplerI0),
        (ASN_TAG_CLASS_CONTEXT | (8 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_keplerI0_constraint_1,
        &ASN_PER_MEMB_KEPLER_I0_CONSTR_10, 0, "keplerI0"},
    {ATF_NOFLAGS, 0, offsetof(struct NavModel_KeplerianSet, keplerOmega0),
        (ASN_TAG_CLASS_CONTEXT | (9 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_keplerOmega0_constraint_1,
        &ASN_PER_MEMB_KEPLER_OMEGA0_CONSTR_11, 0, "keplerOmega0"},
    {ATF_NOFLAGS, 0, offsetof(struct NavModel_KeplerianSet, keplerCrs),
        (ASN_TAG_CLASS_CONTEXT | (10 << 2)),
        -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_keplerCrs_constraint_1,
        &ASN_PER_MEMB_KEPLER_CRS_CONSTR_12, 0, "keplerCrs"},
    {ATF_NOFLAGS, 0, offsetof(struct NavModel_KeplerianSet, keplerCis),
        (ASN_TAG_CLASS_CONTEXT | (11 << 2)),
        -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_keplerCis_constraint_1,
        &ASN_PER_MEMB_KEPLER_CIS_CONSTR_13, 0, "keplerCis"},
    {ATF_NOFLAGS, 0, offsetof(struct NavModel_KeplerianSet, keplerCus),
        (ASN_TAG_CLASS_CONTEXT | (12 << 2)),
        -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_keplerCus_constraint_1,
        &ASN_PER_MEMB_KEPLER_CUS_CONSTR_14, 0, "keplerCus"},
    {ATF_NOFLAGS, 0, offsetof(struct NavModel_KeplerianSet, keplerCrc),
        (ASN_TAG_CLASS_CONTEXT | (13 << 2)),
        -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_keplerCrc_constraint_1,
        &ASN_PER_MEMB_KEPLER_CRC_CONSTR_15, 0, "keplerCrc"},
    {ATF_NOFLAGS, 0, offsetof(struct NavModel_KeplerianSet, keplerCic),
        (ASN_TAG_CLASS_CONTEXT | (14 << 2)),
        -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_keplerCic_constraint_1,
        &ASN_PER_MEMB_KEPLER_CIC_CONSTR_16, 0, "keplerCic"},
    {ATF_NOFLAGS, 0, offsetof(struct NavModel_KeplerianSet, keplerCuc),
        (ASN_TAG_CLASS_CONTEXT | (15 << 2)),
        -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_keplerCuc_constraint_1,
        &ASN_PER_MEMB_KEPLER_CUC_CONSTR_17, 0, "keplerCuc"},
};
static ber_tlv_tag_t asn_DEF_NavModel_KeplerianSet_tags_1[] = {
    (ASN_TAG_CLASS_UNIVERSAL | (16 << 2))};
static asn_TYPE_tag2member_t asn_MAP_NavModel_KeplerianSet_tag2el_1[] = {
    {(ASN_TAG_CLASS_CONTEXT | (0 << 2)), 0, 0, 0}, /* keplerToeLSB at 1254 */
    {(ASN_TAG_CLASS_CONTEXT | (1 << 2)), 1, 0, 0}, /* keplerW at 1255 */
    {(ASN_TAG_CLASS_CONTEXT | (2 << 2)), 2, 0, 0}, /* keplerDeltaN at 1256 */
    {(ASN_TAG_CLASS_CONTEXT | (3 << 2)), 3, 0, 0}, /* keplerM0 at 1257 */
    {(ASN_TAG_CLASS_CONTEXT | (4 << 2)), 4, 0, 0}, /* keplerOmegaDot at 1258 */
    {(ASN_TAG_CLASS_CONTEXT | (5 << 2)), 5, 0, 0}, /* keplerELSB at 1259 */
    {(ASN_TAG_CLASS_CONTEXT | (6 << 2)), 6, 0, 0}, /* keplerIDot at 1260 */
    {(ASN_TAG_CLASS_CONTEXT | (7 << 2)), 7, 0,
        0},                                          /* keplerAPowerHalfLSB at 1261 */
    {(ASN_TAG_CLASS_CONTEXT | (8 << 2)), 8, 0, 0},   /* keplerI0 at 1262 */
    {(ASN_TAG_CLASS_CONTEXT | (9 << 2)), 9, 0, 0},   /* keplerOmega0 at 1263 */
    {(ASN_TAG_CLASS_CONTEXT | (10 << 2)), 10, 0, 0}, /* keplerCrs at 1264 */
    {(ASN_TAG_CLASS_CONTEXT | (11 << 2)), 11, 0, 0}, /* keplerCis at 1265 */
    {(ASN_TAG_CLASS_CONTEXT | (12 << 2)), 12, 0, 0}, /* keplerCus at 1266 */
    {(ASN_TAG_CLASS_CONTEXT | (13 << 2)), 13, 0, 0}, /* keplerCrc at 1267 */
    {(ASN_TAG_CLASS_CONTEXT | (14 << 2)), 14, 0, 0}, /* keplerCic at 1268 */
    {(ASN_TAG_CLASS_CONTEXT | (15 << 2)), 15, 0, 0}  /* keplerCuc at 1269 */
};
static asn_SEQUENCE_specifics_t asn_SPC_NavModel_KeplerianSet_specs_1 = {
    sizeof(struct NavModel_KeplerianSet),
    offsetof(struct NavModel_KeplerianSet, _asn_ctx),
    asn_MAP_NavModel_KeplerianSet_tag2el_1,
    16, /* Count of tags in the map */
    0,
    0,
    0,  /* Optional elements (not needed) */
    -1, /* Start extensions */
    -1  /* Stop extensions */
};
asn_TYPE_descriptor_t asn_DEF_NavModel_KeplerianSet = {
    "NavModel-KeplerianSet",
    "NavModel-KeplerianSet",
    SEQUENCE_free,
    SEQUENCE_print,
    SEQUENCE_constraint,
    SEQUENCE_decode_ber,
    SEQUENCE_encode_der,
    SEQUENCE_decode_xer,
    SEQUENCE_encode_xer,
    SEQUENCE_decode_uper,
    SEQUENCE_encode_uper,
    0, /* Use generic outmost tag fetcher */
    asn_DEF_NavModel_KeplerianSet_tags_1,
    sizeof(asn_DEF_NavModel_KeplerianSet_tags_1) /
        sizeof(asn_DEF_NavModel_KeplerianSet_tags_1[0]), /* 1 */
    asn_DEF_NavModel_KeplerianSet_tags_1,                /* Same as above */
    sizeof(asn_DEF_NavModel_KeplerianSet_tags_1) /
        sizeof(asn_DEF_NavModel_KeplerianSet_tags_1[0]), /* 1 */
    0,                                                   /* No PER visible constraints */
    asn_MBR_NavModel_KeplerianSet_1,
    16,                                    /* Elements count */
    &asn_SPC_NavModel_KeplerianSet_specs_1 /* Additional specs */
};
