/*
 * SPDX-FileCopyrightText: (c) 2003, 2004 Lev Walkin <vlm@lionet.info>. All rights reserved.
 * SPDX-License-Identifier: BSD-1-Clause
 * Generated by asn1c-0.9.22 (http://lionet.info/asn1c)
 * From ASN.1 module "RRLP-Components"
 *     found in "../rrlp-components.asn"
 */

#include "UncompressedEphemeris.h"

static int ephemE_17_constraint(asn_TYPE_descriptor_t *td, const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    unsigned long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const unsigned long *)sptr;

    /* Constraint check succeeded */
    return 0;
}

/*
 * This type is implemented using NativeInteger,
 * so here we adjust the DEF accordingly.
 */
static void ephemE_17_inherit_TYPE_descriptor(asn_TYPE_descriptor_t *td)
{
    td->free_struct = asn_DEF_NativeInteger.free_struct;
    td->print_struct = asn_DEF_NativeInteger.print_struct;
    td->ber_decoder = asn_DEF_NativeInteger.ber_decoder;
    td->der_encoder = asn_DEF_NativeInteger.der_encoder;
    td->xer_decoder = asn_DEF_NativeInteger.xer_decoder;
    td->xer_encoder = asn_DEF_NativeInteger.xer_encoder;
    td->uper_decoder = asn_DEF_NativeInteger.uper_decoder;
    td->uper_encoder = asn_DEF_NativeInteger.uper_encoder;
    if (!td->per_constraints)
        {
            td->per_constraints = asn_DEF_NativeInteger.per_constraints;
        }
    td->elements = asn_DEF_NativeInteger.elements;
    td->elements_count = asn_DEF_NativeInteger.elements_count;
    /* td->specifics      = asn_DEF_NativeInteger.specifics;    // Defined
     * explicitly */
}

static void ephemE_17_free(asn_TYPE_descriptor_t *td, void *struct_ptr,
    int contents_only)
{
    ephemE_17_inherit_TYPE_descriptor(td);
    td->free_struct(td, struct_ptr, contents_only);
}

static int ephemE_17_print(asn_TYPE_descriptor_t *td, const void *struct_ptr,
    int ilevel, asn_app_consume_bytes_f *cb,
    void *app_key)
{
    ephemE_17_inherit_TYPE_descriptor(td);
    return td->print_struct(td, struct_ptr, ilevel, cb, app_key);
}

static asn_dec_rval_t ephemE_17_decode_ber(asn_codec_ctx_t *opt_codec_ctx,
    asn_TYPE_descriptor_t *td,
    void **structure, const void *bufptr,
    size_t size, int tag_mode)
{
    ephemE_17_inherit_TYPE_descriptor(td);
    return td->ber_decoder(opt_codec_ctx, td, structure, bufptr, size,
        tag_mode);
}

static asn_enc_rval_t ephemE_17_encode_der(asn_TYPE_descriptor_t *td,
    void *structure, int tag_mode,
    ber_tlv_tag_t tag,
    asn_app_consume_bytes_f *cb,
    void *app_key)
{
    ephemE_17_inherit_TYPE_descriptor(td);
    return td->der_encoder(td, structure, tag_mode, tag, cb, app_key);
}

static asn_dec_rval_t ephemE_17_decode_xer(asn_codec_ctx_t *opt_codec_ctx,
    asn_TYPE_descriptor_t *td,
    void **structure,
    const char *opt_mname,
    const void *bufptr, size_t size)
{
    ephemE_17_inherit_TYPE_descriptor(td);
    return td->xer_decoder(opt_codec_ctx, td, structure, opt_mname, bufptr,
        size);
}

static asn_enc_rval_t ephemE_17_encode_xer(asn_TYPE_descriptor_t *td,
    void *structure, int ilevel,
    enum xer_encoder_flags_e flags,
    asn_app_consume_bytes_f *cb,
    void *app_key)
{
    ephemE_17_inherit_TYPE_descriptor(td);
    return td->xer_encoder(td, structure, ilevel, flags, cb, app_key);
}

static asn_dec_rval_t ephemE_17_decode_uper(asn_codec_ctx_t *opt_codec_ctx,
    asn_TYPE_descriptor_t *td,
    asn_per_constraints_t *constraints,
    void **structure,
    asn_per_data_t *per_data)
{
    ephemE_17_inherit_TYPE_descriptor(td);
    return td->uper_decoder(opt_codec_ctx, td, constraints, structure,
        per_data);
}

static asn_enc_rval_t ephemE_17_encode_uper(asn_TYPE_descriptor_t *td,
    asn_per_constraints_t *constraints,
    void *structure,
    asn_per_outp_t *per_out)
{
    ephemE_17_inherit_TYPE_descriptor(td);
    return td->uper_encoder(td, constraints, structure, per_out);
}

static int ephemAPowerHalf_19_constraint(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    unsigned long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const unsigned long *)sptr;

    /* Constraint check succeeded */
    return 0;
}

/*
 * This type is implemented using NativeInteger,
 * so here we adjust the DEF accordingly.
 */
static void ephemAPowerHalf_19_inherit_TYPE_descriptor(
    asn_TYPE_descriptor_t *td)
{
    td->free_struct = asn_DEF_NativeInteger.free_struct;
    td->print_struct = asn_DEF_NativeInteger.print_struct;
    td->ber_decoder = asn_DEF_NativeInteger.ber_decoder;
    td->der_encoder = asn_DEF_NativeInteger.der_encoder;
    td->xer_decoder = asn_DEF_NativeInteger.xer_decoder;
    td->xer_encoder = asn_DEF_NativeInteger.xer_encoder;
    td->uper_decoder = asn_DEF_NativeInteger.uper_decoder;
    td->uper_encoder = asn_DEF_NativeInteger.uper_encoder;
    if (!td->per_constraints)
        {
            td->per_constraints = asn_DEF_NativeInteger.per_constraints;
        }
    td->elements = asn_DEF_NativeInteger.elements;
    td->elements_count = asn_DEF_NativeInteger.elements_count;
    /* td->specifics      = asn_DEF_NativeInteger.specifics;    // Defined
     * explicitly */
}

static void ephemAPowerHalf_19_free(asn_TYPE_descriptor_t *td, void *struct_ptr,
    int contents_only)
{
    ephemAPowerHalf_19_inherit_TYPE_descriptor(td);
    td->free_struct(td, struct_ptr, contents_only);
}

static int ephemAPowerHalf_19_print(asn_TYPE_descriptor_t *td,
    const void *struct_ptr, int ilevel,
    asn_app_consume_bytes_f *cb, void *app_key)
{
    ephemAPowerHalf_19_inherit_TYPE_descriptor(td);
    return td->print_struct(td, struct_ptr, ilevel, cb, app_key);
}

static asn_dec_rval_t ephemAPowerHalf_19_decode_ber(
    asn_codec_ctx_t *opt_codec_ctx, asn_TYPE_descriptor_t *td, void **structure,
    const void *bufptr, size_t size, int tag_mode)
{
    ephemAPowerHalf_19_inherit_TYPE_descriptor(td);
    return td->ber_decoder(opt_codec_ctx, td, structure, bufptr, size,
        tag_mode);
}

static asn_enc_rval_t ephemAPowerHalf_19_encode_der(
    asn_TYPE_descriptor_t *td, void *structure, int tag_mode, ber_tlv_tag_t tag,
    asn_app_consume_bytes_f *cb, void *app_key)
{
    ephemAPowerHalf_19_inherit_TYPE_descriptor(td);
    return td->der_encoder(td, structure, tag_mode, tag, cb, app_key);
}

static asn_dec_rval_t ephemAPowerHalf_19_decode_xer(
    asn_codec_ctx_t *opt_codec_ctx, asn_TYPE_descriptor_t *td, void **structure,
    const char *opt_mname, const void *bufptr, size_t size)
{
    ephemAPowerHalf_19_inherit_TYPE_descriptor(td);
    return td->xer_decoder(opt_codec_ctx, td, structure, opt_mname, bufptr,
        size);
}

static asn_enc_rval_t ephemAPowerHalf_19_encode_xer(
    asn_TYPE_descriptor_t *td, void *structure, int ilevel,
    enum xer_encoder_flags_e flags, asn_app_consume_bytes_f *cb, void *app_key)
{
    ephemAPowerHalf_19_inherit_TYPE_descriptor(td);
    return td->xer_encoder(td, structure, ilevel, flags, cb, app_key);
}

static asn_dec_rval_t ephemAPowerHalf_19_decode_uper(
    asn_codec_ctx_t *opt_codec_ctx, asn_TYPE_descriptor_t *td,
    asn_per_constraints_t *constraints, void **structure,
    asn_per_data_t *per_data)
{
    ephemAPowerHalf_19_inherit_TYPE_descriptor(td);
    return td->uper_decoder(opt_codec_ctx, td, constraints, structure,
        per_data);
}

static asn_enc_rval_t ephemAPowerHalf_19_encode_uper(
    asn_TYPE_descriptor_t *td, asn_per_constraints_t *constraints,
    void *structure, asn_per_outp_t *per_out)
{
    ephemAPowerHalf_19_inherit_TYPE_descriptor(td);
    return td->uper_encoder(td, constraints, structure, per_out);
}

static int memb_ephemCodeOnL2_constraint_1(
    asn_TYPE_descriptor_t *td, const void *sptr,
    asn_app_constraint_failed_f *ctfailcb, void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= 0 && value <= 3))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ephemURA_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= 0 && value <= 15))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ephemSVhealth_constraint_1(
    asn_TYPE_descriptor_t *td, const void *sptr,
    asn_app_constraint_failed_f *ctfailcb, void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= 0 && value <= 63))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ephemIODC_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= 0 && value <= 1023))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ephemL2Pflag_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= 0 && value <= 1))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ephemTgd_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= -128 && value <= 127))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ephemToc_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= 0 && value <= 37799))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ephemAF2_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= -128 && value <= 127))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ephemAF1_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= -32768 && value <= 32767))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ephemAF0_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= -2097152 && value <= 2097151))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ephemCrs_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= -32768 && value <= 32767))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ephemDeltaN_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= -32768 && value <= 32767))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ephemM0_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= (-2147483647L - 1) && value <= 2147483647))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ephemCuc_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= -32768 && value <= 32767))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ephemE_constraint_1(asn_TYPE_descriptor_t *td, const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    unsigned long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const unsigned long *)sptr;

    /* Constraint check succeeded */
    return 0;
}

static int memb_ephemCus_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= -32768 && value <= 32767))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ephemAPowerHalf_constraint_1(
    asn_TYPE_descriptor_t *td, const void *sptr,
    asn_app_constraint_failed_f *ctfailcb, void *app_key)
{
    unsigned long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const unsigned long *)sptr;

    /* Constraint check succeeded */
    return 0;
}

static int memb_ephemToe_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= 0 && value <= 37799))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ephemFitFlag_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= 0 && value <= 1))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ephemAODA_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= 0 && value <= 31))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ephemCic_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= -32768 && value <= 32767))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ephemOmegaA0_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= (-2147483647L - 1) && value <= 2147483647))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ephemCis_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= -32768 && value <= 32767))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ephemI0_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= (-2147483647L - 1) && value <= 2147483647))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ephemCrc_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= -32768 && value <= 32767))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ephemW_constraint_1(asn_TYPE_descriptor_t *td, const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= (-2147483647L - 1) && value <= 2147483647))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ephemOmegaADot_constraint_1(
    asn_TYPE_descriptor_t *td, const void *sptr,
    asn_app_constraint_failed_f *ctfailcb, void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= -8388608 && value <= 8388607))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ephemIDot_constraint_1(asn_TYPE_descriptor_t *td,
    const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= -8192 && value <= 8191))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static asn_per_constraints_t ASN_PER_TYPE_EPHEM_E_CONSTR_17 = {
    {APC_CONSTRAINED, 32, -1, 0, 4294967295} /* (0..4294967295) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_TYPE_EPHEM_A_POWER_HALF_CONSTR_19 = {
    {APC_CONSTRAINED, 32, -1, 0, 4294967295} /* (0..4294967295) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_CODE_ON_L2_CONSTR_2 = {
    {APC_CONSTRAINED, 2, 2, 0, 3} /* (0..3) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_URA_CONSTR_3 = {
    {APC_CONSTRAINED, 4, 4, 0, 15} /* (0..15) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_S_VHEALTH_CONSTR_4 = {
    {APC_CONSTRAINED, 6, 6, 0, 63} /* (0..63) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_IODC_CONSTR_5 = {
    {APC_CONSTRAINED, 10, 10, 0, 1023} /* (0..1023) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_L2_PFLAG_CONSTR_6 = {
    {APC_CONSTRAINED, 1, 1, 0, 1} /* (0..1) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_TGD_CONSTR_8 = {
    {APC_CONSTRAINED, 8, 8, -128, 127} /* (-128..127) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_TOC_CONSTR_9 = {
    {APC_CONSTRAINED, 16, 16, 0, 37799} /* (0..37799) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_A_F2_CONSTR_10 = {
    {APC_CONSTRAINED, 8, 8, -128, 127} /* (-128..127) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_A_F1_CONSTR_11 = {
    {APC_CONSTRAINED, 16, 16, -32768, 32767} /* (-32768..32767) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_A_F0_CONSTR_12 = {
    {APC_CONSTRAINED, 22, -1, -2097152, 2097151} /* (-2097152..2097151) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_CRS_CONSTR_13 = {
    {APC_CONSTRAINED, 16, 16, -32768, 32767} /* (-32768..32767) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_DELTA_N_CONSTR_14 = {
    {APC_CONSTRAINED, 16, 16, -32768, 32767} /* (-32768..32767) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_M0_CONSTR_15 = {
    {APC_CONSTRAINED, 32, -1, (-2147483647L - 1),
        2147483647} /* (-2147483648..2147483647) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_CUC_CONSTR_16 = {
    {APC_CONSTRAINED, 16, 16, -32768, 32767} /* (-32768..32767) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_E_CONSTR_17 = {
    {APC_CONSTRAINED, 32, -1, 0, 4294967295} /* (0..4294967295) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_CUS_CONSTR_18 = {
    {APC_CONSTRAINED, 16, 16, -32768, 32767} /* (-32768..32767) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_A_POWER_HALF_CONSTR_19 = {
    {APC_CONSTRAINED, 32, -1, 0, 4294967295} /* (0..4294967295) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_TOE_CONSTR_20 = {
    {APC_CONSTRAINED, 16, 16, 0, 37799} /* (0..37799) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_FIT_FLAG_CONSTR_21 = {
    {APC_CONSTRAINED, 1, 1, 0, 1} /* (0..1) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_AODA_CONSTR_22 = {
    {APC_CONSTRAINED, 5, 5, 0, 31} /* (0..31) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_CIC_CONSTR_23 = {
    {APC_CONSTRAINED, 16, 16, -32768, 32767} /* (-32768..32767) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_OMEGA_A0_CONSTR_24 = {
    {APC_CONSTRAINED, 32, -1, (-2147483647L - 1),
        2147483647} /* (-2147483648..2147483647) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_CIS_CONSTR_25 = {
    {APC_CONSTRAINED, 16, 16, -32768, 32767} /* (-32768..32767) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_I0_CONSTR_26 = {
    {APC_CONSTRAINED, 32, -1, (-2147483647L - 1),
        2147483647} /* (-2147483648..2147483647) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_CRC_CONSTR_27 = {
    {APC_CONSTRAINED, 16, 16, -32768, 32767} /* (-32768..32767) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_W_CONSTR_28 = {
    {APC_CONSTRAINED, 32, -1, (-2147483647L - 1),
        2147483647} /* (-2147483648..2147483647) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_OMEGA_A_DOT_CONSTR_29 = {
    {APC_CONSTRAINED, 24, -1, -8388608, 8388607} /* (-8388608..8388607) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_EPHEM_I_DOT_CONSTR_30 = {
    {APC_CONSTRAINED, 14, 14, -8192, 8191} /* (-8192..8191) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_INTEGER_specifics_t asn_SPC_ephemE_specs_17 = {
    0, 0, 0, 0, 0, 0, /* Native long size */
    1                 /* Unsigned representation */
};
static ber_tlv_tag_t asn_DEF_ephemE_tags_17[] = {
    (ASN_TAG_CLASS_CONTEXT | (15 << 2)), (ASN_TAG_CLASS_UNIVERSAL | (2 << 2))};
static /* Use -fall-defs-global to expose */
    asn_TYPE_descriptor_t asn_DEF_ephemE_17 = {
        "ephemE",
        "ephemE",
        ephemE_17_free,
        ephemE_17_print,
        ephemE_17_constraint,
        ephemE_17_decode_ber,
        ephemE_17_encode_der,
        ephemE_17_decode_xer,
        ephemE_17_encode_xer,
        ephemE_17_decode_uper,
        ephemE_17_encode_uper,
        0, /* Use generic outmost tag fetcher */
        asn_DEF_ephemE_tags_17,
        sizeof(asn_DEF_ephemE_tags_17) / sizeof(asn_DEF_ephemE_tags_17[0]) -
            1,                  /* 1 */
        asn_DEF_ephemE_tags_17, /* Same as above */
        sizeof(asn_DEF_ephemE_tags_17) /
            sizeof(asn_DEF_ephemE_tags_17[0]), /* 2 */
        &ASN_PER_TYPE_EPHEM_E_CONSTR_17,
        0,
        0,                       /* No members */
        &asn_SPC_ephemE_specs_17 /* Additional specs */
};

static asn_INTEGER_specifics_t asn_SPC_ephemAPowerHalf_specs_19 = {
    0, 0, 0, 0, 0, 0, /* Native long size */
    1                 /* Unsigned representation */
};
static ber_tlv_tag_t asn_DEF_ephemAPowerHalf_tags_19[] = {
    (ASN_TAG_CLASS_CONTEXT | (17 << 2)), (ASN_TAG_CLASS_UNIVERSAL | (2 << 2))};
static /* Use -fall-defs-global to expose */
    asn_TYPE_descriptor_t asn_DEF_ephemAPowerHalf_19 = {
        "ephemAPowerHalf",
        "ephemAPowerHalf",
        ephemAPowerHalf_19_free,
        ephemAPowerHalf_19_print,
        ephemAPowerHalf_19_constraint,
        ephemAPowerHalf_19_decode_ber,
        ephemAPowerHalf_19_encode_der,
        ephemAPowerHalf_19_decode_xer,
        ephemAPowerHalf_19_encode_xer,
        ephemAPowerHalf_19_decode_uper,
        ephemAPowerHalf_19_encode_uper,
        0, /* Use generic outmost tag fetcher */
        asn_DEF_ephemAPowerHalf_tags_19,
        sizeof(asn_DEF_ephemAPowerHalf_tags_19) /
                sizeof(asn_DEF_ephemAPowerHalf_tags_19[0]) -
            1,                           /* 1 */
        asn_DEF_ephemAPowerHalf_tags_19, /* Same as above */
        sizeof(asn_DEF_ephemAPowerHalf_tags_19) /
            sizeof(asn_DEF_ephemAPowerHalf_tags_19[0]), /* 2 */
        &ASN_PER_TYPE_EPHEM_A_POWER_HALF_CONSTR_19,
        0,
        0,                                /* No members */
        &asn_SPC_ephemAPowerHalf_specs_19 /* Additional specs */
};

static asn_TYPE_member_t asn_MBR_UncompressedEphemeris_1[] = {
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemCodeOnL2),
        (ASN_TAG_CLASS_CONTEXT | (0 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ephemCodeOnL2_constraint_1,
        &ASN_PER_MEMB_EPHEM_CODE_ON_L2_CONSTR_2, 0, "ephemCodeOnL2"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemURA),
        (ASN_TAG_CLASS_CONTEXT | (1 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ephemURA_constraint_1,
        &ASN_PER_MEMB_EPHEM_URA_CONSTR_3, 0, "ephemURA"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemSVhealth),
        (ASN_TAG_CLASS_CONTEXT | (2 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ephemSVhealth_constraint_1,
        &ASN_PER_MEMB_EPHEM_S_VHEALTH_CONSTR_4, 0, "ephemSVhealth"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemIODC),
        (ASN_TAG_CLASS_CONTEXT | (3 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ephemIODC_constraint_1,
        &ASN_PER_MEMB_EPHEM_IODC_CONSTR_5, 0, "ephemIODC"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemL2Pflag),
        (ASN_TAG_CLASS_CONTEXT | (4 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ephemL2Pflag_constraint_1,
        &ASN_PER_MEMB_EPHEM_L2_PFLAG_CONSTR_6, 0, "ephemL2Pflag"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemSF1Rsvd),
        (ASN_TAG_CLASS_CONTEXT | (5 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_EphemerisSubframe1Reserved,
        0, /* Defer constraints checking to the member type */
        0, /* No PER visible constraints */
        0, "ephemSF1Rsvd"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemTgd),
        (ASN_TAG_CLASS_CONTEXT | (6 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ephemTgd_constraint_1,
        &ASN_PER_MEMB_EPHEM_TGD_CONSTR_8, 0, "ephemTgd"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemToc),
        (ASN_TAG_CLASS_CONTEXT | (7 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ephemToc_constraint_1,
        &ASN_PER_MEMB_EPHEM_TOC_CONSTR_9, 0, "ephemToc"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemAF2),
        (ASN_TAG_CLASS_CONTEXT | (8 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ephemAF2_constraint_1,
        &ASN_PER_MEMB_EPHEM_A_F2_CONSTR_10, 0, "ephemAF2"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemAF1),
        (ASN_TAG_CLASS_CONTEXT | (9 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ephemAF1_constraint_1,
        &ASN_PER_MEMB_EPHEM_A_F1_CONSTR_11, 0, "ephemAF1"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemAF0),
        (ASN_TAG_CLASS_CONTEXT | (10 << 2)),
        -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ephemAF0_constraint_1,
        &ASN_PER_MEMB_EPHEM_A_F0_CONSTR_12, 0, "ephemAF0"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemCrs),
        (ASN_TAG_CLASS_CONTEXT | (11 << 2)),
        -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ephemCrs_constraint_1,
        &ASN_PER_MEMB_EPHEM_CRS_CONSTR_13, 0, "ephemCrs"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemDeltaN),
        (ASN_TAG_CLASS_CONTEXT | (12 << 2)),
        -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ephemDeltaN_constraint_1,
        &ASN_PER_MEMB_EPHEM_DELTA_N_CONSTR_14, 0, "ephemDeltaN"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemM0),
        (ASN_TAG_CLASS_CONTEXT | (13 << 2)),
        -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ephemM0_constraint_1,
        &ASN_PER_MEMB_EPHEM_M0_CONSTR_15, 0, "ephemM0"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemCuc),
        (ASN_TAG_CLASS_CONTEXT | (14 << 2)),
        -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ephemCuc_constraint_1,
        &ASN_PER_MEMB_EPHEM_CUC_CONSTR_16, 0, "ephemCuc"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemE),
        (ASN_TAG_CLASS_CONTEXT | (15 << 2)),
        -1, /* IMPLICIT tag at current level */
        &asn_DEF_ephemE_17, memb_ephemE_constraint_1,
        &ASN_PER_MEMB_EPHEM_E_CONSTR_17, 0, "ephemE"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemCus),
        (ASN_TAG_CLASS_CONTEXT | (16 << 2)),
        -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ephemCus_constraint_1,
        &ASN_PER_MEMB_EPHEM_CUS_CONSTR_18, 0, "ephemCus"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemAPowerHalf),
        (ASN_TAG_CLASS_CONTEXT | (17 << 2)),
        -1, /* IMPLICIT tag at current level */
        &asn_DEF_ephemAPowerHalf_19, memb_ephemAPowerHalf_constraint_1,
        &ASN_PER_MEMB_EPHEM_A_POWER_HALF_CONSTR_19, 0, "ephemAPowerHalf"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemToe),
        (ASN_TAG_CLASS_CONTEXT | (18 << 2)),
        -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ephemToe_constraint_1,
        &ASN_PER_MEMB_EPHEM_TOE_CONSTR_20, 0, "ephemToe"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemFitFlag),
        (ASN_TAG_CLASS_CONTEXT | (19 << 2)),
        -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ephemFitFlag_constraint_1,
        &ASN_PER_MEMB_EPHEM_FIT_FLAG_CONSTR_21, 0, "ephemFitFlag"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemAODA),
        (ASN_TAG_CLASS_CONTEXT | (20 << 2)),
        -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ephemAODA_constraint_1,
        &ASN_PER_MEMB_EPHEM_AODA_CONSTR_22, 0, "ephemAODA"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemCic),
        (ASN_TAG_CLASS_CONTEXT | (21 << 2)),
        -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ephemCic_constraint_1,
        &ASN_PER_MEMB_EPHEM_CIC_CONSTR_23, 0, "ephemCic"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemOmegaA0),
        (ASN_TAG_CLASS_CONTEXT | (22 << 2)),
        -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ephemOmegaA0_constraint_1,
        &ASN_PER_MEMB_EPHEM_OMEGA_A0_CONSTR_24, 0, "ephemOmegaA0"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemCis),
        (ASN_TAG_CLASS_CONTEXT | (23 << 2)),
        -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ephemCis_constraint_1,
        &ASN_PER_MEMB_EPHEM_CIS_CONSTR_25, 0, "ephemCis"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemI0),
        (ASN_TAG_CLASS_CONTEXT | (24 << 2)),
        -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ephemI0_constraint_1,
        &ASN_PER_MEMB_EPHEM_I0_CONSTR_26, 0, "ephemI0"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemCrc),
        (ASN_TAG_CLASS_CONTEXT | (25 << 2)),
        -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ephemCrc_constraint_1,
        &ASN_PER_MEMB_EPHEM_CRC_CONSTR_27, 0, "ephemCrc"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemW),
        (ASN_TAG_CLASS_CONTEXT | (26 << 2)),
        -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ephemW_constraint_1,
        &ASN_PER_MEMB_EPHEM_W_CONSTR_28, 0, "ephemW"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemOmegaADot),
        (ASN_TAG_CLASS_CONTEXT | (27 << 2)),
        -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ephemOmegaADot_constraint_1,
        &ASN_PER_MEMB_EPHEM_OMEGA_A_DOT_CONSTR_29, 0, "ephemOmegaADot"},
    {ATF_NOFLAGS, 0, offsetof(struct UncompressedEphemeris, ephemIDot),
        (ASN_TAG_CLASS_CONTEXT | (28 << 2)),
        -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ephemIDot_constraint_1,
        &ASN_PER_MEMB_EPHEM_I_DOT_CONSTR_30, 0, "ephemIDot"},
};
static ber_tlv_tag_t asn_DEF_UncompressedEphemeris_tags_1[] = {
    (ASN_TAG_CLASS_UNIVERSAL | (16 << 2))};
static asn_TYPE_tag2member_t asn_MAP_UncompressedEphemeris_tag2el_1[] = {
    {(ASN_TAG_CLASS_CONTEXT | (0 << 2)), 0, 0, 0},   /* ephemCodeOnL2 at 722 */
    {(ASN_TAG_CLASS_CONTEXT | (1 << 2)), 1, 0, 0},   /* ephemURA at 723 */
    {(ASN_TAG_CLASS_CONTEXT | (2 << 2)), 2, 0, 0},   /* ephemSVhealth at 724 */
    {(ASN_TAG_CLASS_CONTEXT | (3 << 2)), 3, 0, 0},   /* ephemIODC at 725 */
    {(ASN_TAG_CLASS_CONTEXT | (4 << 2)), 4, 0, 0},   /* ephemL2Pflag at 726 */
    {(ASN_TAG_CLASS_CONTEXT | (5 << 2)), 5, 0, 0},   /* ephemSF1Rsvd at 727 */
    {(ASN_TAG_CLASS_CONTEXT | (6 << 2)), 6, 0, 0},   /* ephemTgd at 728 */
    {(ASN_TAG_CLASS_CONTEXT | (7 << 2)), 7, 0, 0},   /* ephemToc at 729 */
    {(ASN_TAG_CLASS_CONTEXT | (8 << 2)), 8, 0, 0},   /* ephemAF2 at 730 */
    {(ASN_TAG_CLASS_CONTEXT | (9 << 2)), 9, 0, 0},   /* ephemAF1 at 731 */
    {(ASN_TAG_CLASS_CONTEXT | (10 << 2)), 10, 0, 0}, /* ephemAF0 at 732 */
    {(ASN_TAG_CLASS_CONTEXT | (11 << 2)), 11, 0, 0}, /* ephemCrs at 733 */
    {(ASN_TAG_CLASS_CONTEXT | (12 << 2)), 12, 0, 0}, /* ephemDeltaN at 734 */
    {(ASN_TAG_CLASS_CONTEXT | (13 << 2)), 13, 0, 0}, /* ephemM0 at 735 */
    {(ASN_TAG_CLASS_CONTEXT | (14 << 2)), 14, 0, 0}, /* ephemCuc at 736 */
    {(ASN_TAG_CLASS_CONTEXT | (15 << 2)), 15, 0, 0}, /* ephemE at 737 */
    {(ASN_TAG_CLASS_CONTEXT | (16 << 2)), 16, 0, 0}, /* ephemCus at 738 */
    {(ASN_TAG_CLASS_CONTEXT | (17 << 2)), 17, 0,
        0},                                          /* ephemAPowerHalf at 739 */
    {(ASN_TAG_CLASS_CONTEXT | (18 << 2)), 18, 0, 0}, /* ephemToe at 740 */
    {(ASN_TAG_CLASS_CONTEXT | (19 << 2)), 19, 0, 0}, /* ephemFitFlag at 741 */
    {(ASN_TAG_CLASS_CONTEXT | (20 << 2)), 20, 0, 0}, /* ephemAODA at 742 */
    {(ASN_TAG_CLASS_CONTEXT | (21 << 2)), 21, 0, 0}, /* ephemCic at 743 */
    {(ASN_TAG_CLASS_CONTEXT | (22 << 2)), 22, 0, 0}, /* ephemOmegaA0 at 744 */
    {(ASN_TAG_CLASS_CONTEXT | (23 << 2)), 23, 0, 0}, /* ephemCis at 745 */
    {(ASN_TAG_CLASS_CONTEXT | (24 << 2)), 24, 0, 0}, /* ephemI0 at 746 */
    {(ASN_TAG_CLASS_CONTEXT | (25 << 2)), 25, 0, 0}, /* ephemCrc at 747 */
    {(ASN_TAG_CLASS_CONTEXT | (26 << 2)), 26, 0, 0}, /* ephemW at 748 */
    {(ASN_TAG_CLASS_CONTEXT | (27 << 2)), 27, 0, 0}, /* ephemOmegaADot at 749 */
    {(ASN_TAG_CLASS_CONTEXT | (28 << 2)), 28, 0, 0}  /* ephemIDot at 750 */
};
static asn_SEQUENCE_specifics_t asn_SPC_UncompressedEphemeris_specs_1 = {
    sizeof(struct UncompressedEphemeris),
    offsetof(struct UncompressedEphemeris, _asn_ctx),
    asn_MAP_UncompressedEphemeris_tag2el_1,
    29, /* Count of tags in the map */
    0,
    0,
    0,  /* Optional elements (not needed) */
    -1, /* Start extensions */
    -1  /* Stop extensions */
};
asn_TYPE_descriptor_t asn_DEF_UncompressedEphemeris = {
    "UncompressedEphemeris",
    "UncompressedEphemeris",
    SEQUENCE_free,
    SEQUENCE_print,
    SEQUENCE_constraint,
    SEQUENCE_decode_ber,
    SEQUENCE_encode_der,
    SEQUENCE_decode_xer,
    SEQUENCE_encode_xer,
    SEQUENCE_decode_uper,
    SEQUENCE_encode_uper,
    0, /* Use generic outmost tag fetcher */
    asn_DEF_UncompressedEphemeris_tags_1,
    sizeof(asn_DEF_UncompressedEphemeris_tags_1) /
        sizeof(asn_DEF_UncompressedEphemeris_tags_1[0]), /* 1 */
    asn_DEF_UncompressedEphemeris_tags_1,                /* Same as above */
    sizeof(asn_DEF_UncompressedEphemeris_tags_1) /
        sizeof(asn_DEF_UncompressedEphemeris_tags_1[0]), /* 1 */
    0,                                                   /* No PER visible constraints */
    asn_MBR_UncompressedEphemeris_1,
    29,                                    /* Elements count */
    &asn_SPC_UncompressedEphemeris_specs_1 /* Additional specs */
};
