/*
 * SPDX-FileCopyrightText: (c) 2003, 2004 Lev Walkin <vlm@lionet.info>. All rights reserved.
 * SPDX-License-Identifier: BSD-1-Clause
 * Generated by asn1c-0.9.22 (http://lionet.info/asn1c)
 * From ASN.1 module "RRLP-Components"
 *     found in "../rrlp-components.asn"
 */

#include "GANSSIonosphereModel.h"

static int memb_ai0_constraint_1(asn_TYPE_descriptor_t *td, const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= 0 && value <= 4095))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ai1_constraint_1(asn_TYPE_descriptor_t *td, const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= 0 && value <= 4095))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static int memb_ai2_constraint_1(asn_TYPE_descriptor_t *td, const void *sptr,
    asn_app_constraint_failed_f *ctfailcb,
    void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= 0 && value <= 4095))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static asn_per_constraints_t ASN_PER_MEMB_AI0_CONSTR_2 = {
    {APC_CONSTRAINED, 12, 12, 0, 4095} /* (0..4095) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_AI1_CONSTR_3 = {
    {APC_CONSTRAINED, 12, 12, 0, 4095} /* (0..4095) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_per_constraints_t ASN_PER_MEMB_AI2_CONSTR_4 = {
    {APC_CONSTRAINED, 12, 12, 0, 4095} /* (0..4095) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_TYPE_member_t asn_MBR_GANSSIonosphereModel_1[] = {
    {ATF_NOFLAGS, 0, offsetof(struct GANSSIonosphereModel, ai0),
        (ASN_TAG_CLASS_CONTEXT | (0 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ai0_constraint_1, &ASN_PER_MEMB_AI0_CONSTR_2,
        0, "ai0"},
    {ATF_NOFLAGS, 0, offsetof(struct GANSSIonosphereModel, ai1),
        (ASN_TAG_CLASS_CONTEXT | (1 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ai1_constraint_1, &ASN_PER_MEMB_AI1_CONSTR_3,
        0, "ai1"},
    {ATF_NOFLAGS, 0, offsetof(struct GANSSIonosphereModel, ai2),
        (ASN_TAG_CLASS_CONTEXT | (2 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_ai2_constraint_1, &ASN_PER_MEMB_AI2_CONSTR_4,
        0, "ai2"},
};
static ber_tlv_tag_t asn_DEF_GANSSIonosphereModel_tags_1[] = {
    (ASN_TAG_CLASS_UNIVERSAL | (16 << 2))};
static asn_TYPE_tag2member_t asn_MAP_GANSSIonosphereModel_tag2el_1[] = {
    {(ASN_TAG_CLASS_CONTEXT | (0 << 2)), 0, 0, 0}, /* ai0 at 1145 */
    {(ASN_TAG_CLASS_CONTEXT | (1 << 2)), 1, 0, 0}, /* ai1 at 1146 */
    {(ASN_TAG_CLASS_CONTEXT | (2 << 2)), 2, 0, 0}  /* ai2 at 1147 */
};
static asn_SEQUENCE_specifics_t asn_SPC_GANSSIonosphereModel_specs_1 = {
    sizeof(struct GANSSIonosphereModel),
    offsetof(struct GANSSIonosphereModel, _asn_ctx),
    asn_MAP_GANSSIonosphereModel_tag2el_1,
    3, /* Count of tags in the map */
    0,
    0,
    0,  /* Optional elements (not needed) */
    -1, /* Start extensions */
    -1  /* Stop extensions */
};
asn_TYPE_descriptor_t asn_DEF_GANSSIonosphereModel = {
    "GANSSIonosphereModel",
    "GANSSIonosphereModel",
    SEQUENCE_free,
    SEQUENCE_print,
    SEQUENCE_constraint,
    SEQUENCE_decode_ber,
    SEQUENCE_encode_der,
    SEQUENCE_decode_xer,
    SEQUENCE_encode_xer,
    SEQUENCE_decode_uper,
    SEQUENCE_encode_uper,
    0, /* Use generic outmost tag fetcher */
    asn_DEF_GANSSIonosphereModel_tags_1,
    sizeof(asn_DEF_GANSSIonosphereModel_tags_1) /
        sizeof(asn_DEF_GANSSIonosphereModel_tags_1[0]), /* 1 */
    asn_DEF_GANSSIonosphereModel_tags_1,                /* Same as above */
    sizeof(asn_DEF_GANSSIonosphereModel_tags_1) /
        sizeof(asn_DEF_GANSSIonosphereModel_tags_1[0]), /* 1 */
    0,                                                  /* No PER visible constraints */
    asn_MBR_GANSSIonosphereModel_1,
    3,                                    /* Elements count */
    &asn_SPC_GANSSIonosphereModel_specs_1 /* Additional specs */
};
