/*
 * SPDX-FileCopyrightText: (c) 2003, 2004 Lev Walkin <vlm@lionet.info>. All rights reserved.
 * SPDX-License-Identifier: BSD-1-Clause
 * Generated by asn1c-0.9.22 (http://lionet.info/asn1c)
 * From ASN.1 module "ULP"
 *     found in "../supl-ulp.asn"
 */

#ifndef _ULP_PDU_H
#define _ULP_PDU_H

#include <asn_application.h>

/* Including external dependencies */
#include "SessionID.h"
#include "UlpMessage.h"
#include "Version.h"
#include <NativeInteger.h>
#include <constr_SEQUENCE.h>

#ifdef __cplusplus
extern "C"
{
#endif

    /* ULP-PDU */
    typedef struct ULP_PDU
    {
        long length;
        Version_t version;
        SessionID_t sessionID;
        UlpMessage_t message;

        /* Context for parsing across buffer boundaries */
        asn_struct_ctx_t _asn_ctx;
    } ULP_PDU_t;

    /* Implementation */
    extern asn_TYPE_descriptor_t asn_DEF_ULP_PDU;

#ifdef __cplusplus
}
#endif

#endif /* _ULP_PDU_H_ */
#include <asn_internal.h>
