/*
 * SPDX-FileCopyrightText: (c) 2003, 2004 Lev Walkin <vlm@lionet.info>. All rights reserved.
 * SPDX-License-Identifier: BSD-1-Clause
 * Generated by asn1c-0.9.22 (http://lionet.info/asn1c)
 * From ASN.1 module "ULP-Components"
 *     found in "../supl-common.asn"
 */

#ifndef _FrequencyInfo_H
#define _FrequencyInfo_H

#include <asn_application.h>

/* Including external dependencies */
#include "FrequencyInfoFDD.h"
#include "FrequencyInfoTDD.h"
#include <constr_CHOICE.h>
#include <constr_SEQUENCE.h>

#ifdef __cplusplus
extern "C"
{
#endif

    /* Dependencies */
    typedef enum fmodeSpecificInfo_PR
    {
        fmodeSpecificInfo_PR_NOTHING, /* No components present */
        fmodeSpecificInfo_PR_fdd,
        fmodeSpecificInfo_PR_tdd,
        /* Extensions may appear below */
    } fmodeSpecificInfo_PR;

    /* FrequencyInfo */
    typedef struct FrequencyInfo
    {
        struct fmodeSpecificInfo
        {
            fmodeSpecificInfo_PR present;
            union FrequencyInfo__fmodeSpecificInfo_u
            {
                FrequencyInfoFDD_t fdd;
                FrequencyInfoTDD_t tdd;
                /*
                 * This type is extensible,
                 * possible extensions are below.
                 */
            } choice;

            /* Context for parsing across buffer boundaries */
            asn_struct_ctx_t _asn_ctx;
        } fmodeSpecificInfo;
        /*
         * This type is extensible,
         * possible extensions are below.
         */

        /* Context for parsing across buffer boundaries */
        asn_struct_ctx_t _asn_ctx;
    } FrequencyInfo_t;

    /* Implementation */
    extern asn_TYPE_descriptor_t asn_DEF_FrequencyInfo;

#ifdef __cplusplus
}
#endif

#endif /* _FrequencyInfo_H_ */
#include <asn_internal.h>
