/*-
 * SPDX-FileCopyrightText: (c) 2003, 2004 Lev Walkin <vlm@lionet.info>. All rights reserved.
 * SPDX-License-Identifier: BSD-1-Clause
 */
#ifndef _CONSTR_SEQUENCE_H
#define _CONSTR_SEQUENCE_H

#include <asn_application.h>

#ifdef __cplusplus
extern "C"
{
#endif

    typedef struct asn_SEQUENCE_specifics_s
    {
        /*
         * Target structure description.
         */
        int struct_size; /* Size of the target structure. */
        int ctx_offset;  /* Offset of the asn_struct_ctx_t member */

        /*
         * Tags to members mapping table (sorted).
         */
        asn_TYPE_tag2member_t *tag2el;
        int tag2el_count;

        /*
         * Optional members of the extensions root (roms) or additions (aoms).
         * Meaningful for PER.
         */
        int *oms;       /* Optional MemberS */
        int roms_count; /* Root optional members count */
        int aoms_count; /* Additions optional members count */

        /*
         * Description of an extensions group.
         */
        int ext_after;  /* Extensions start after this member */
        int ext_before; /* Extensions stop before this member */
    } asn_SEQUENCE_specifics_t;

    /*
     * A set specialized functions dealing with the SEQUENCE type.
     */
    asn_struct_free_f SEQUENCE_free;
    asn_struct_print_f SEQUENCE_print;
    asn_constr_check_f SEQUENCE_constraint;
    ber_type_decoder_f SEQUENCE_decode_ber;
    der_type_encoder_f SEQUENCE_encode_der;
    xer_type_decoder_f SEQUENCE_decode_xer;
    xer_type_encoder_f SEQUENCE_encode_xer;
    per_type_decoder_f SEQUENCE_decode_uper;
    per_type_encoder_f SEQUENCE_encode_uper;

#ifdef __cplusplus
}
#endif

#endif /* _CONSTR_SEQUENCE_H_ */
