#!/usr/bin/env python
#
# GNSS-SDR is a Global Navigation Satellite System software-defined receiver.
# This file is part of GNSS-SDR.
#
# Copyright (C) 2010-2020  (see AUTHORS file for a list of contributors)
# SPDX-License-Identifier: GPL-3.0-or-later

from .cfg import volk_gnsssdr_modtool_config
from .volk_gnsssdr_modtool_generate import volk_gnsssdr_modtool
