/*-
 * SPDX-FileCopyrightText: (c) 2003, 2004 Lev Walkin <vlm@lionet.info>. All rights reserved.
 * SPDX-License-Identifier: BSD-1-Clause
 */
/*
 * This type differs from the standard INTEGER in that it is modelled using
 * the fixed machine type (long, int, short), so it can hold only values of
 * limited length. There is no type (i.e., NativeInteger_t, any integer type
 * will do).
 * This type may be used when integer range is limited by subtype constraints.
 */
#ifndef _NativeInteger_H
#define _NativeInteger_H

#include <INTEGER.h>
#include <asn_application.h>

#ifdef __cplusplus
extern "C"
{
#endif

    extern asn_TYPE_descriptor_t asn_DEF_NativeInteger;

    asn_struct_free_f NativeInteger_free;
    asn_struct_print_f NativeInteger_print;
    ber_type_decoder_f NativeInteger_decode_ber;
    der_type_encoder_f NativeInteger_encode_der;
    xer_type_decoder_f NativeInteger_decode_xer;
    xer_type_encoder_f NativeInteger_encode_xer;
    per_type_decoder_f NativeInteger_decode_uper;
    per_type_encoder_f NativeInteger_encode_uper;

#ifdef __cplusplus
}
#endif

#endif /* _NativeInteger_H_ */
