/*
 * SPDX-FileCopyrightText: (c) 2003, 2004 Lev Walkin <vlm@lionet.info>. All rights reserved.
 * SPDX-License-Identifier: BSD-1-Clause
 * Generated by asn1c-0.9.22 (http://lionet.info/asn1c)
 * From ASN.1 module "RRLP-Components"
 *     found in "../rrlp-components.asn"
 */

#ifndef _GANSS_MsrSetElement_H
#define _GANSS_MsrSetElement_H

#include <asn_application.h>

/* Including external dependencies */
#include "GANSSTODUncertainty.h"
#include "GANSSTODm.h"
#include "SeqOfGANSS-SgnTypeElement.h"
#include <NativeInteger.h>
#include <constr_SEQUENCE.h>

#ifdef __cplusplus
extern "C"
{
#endif

    /* Forward declarations */
    struct ReferenceFrame;

    /* GANSS-MsrSetElement */
    typedef struct GANSS_MsrSetElement
    {
        struct ReferenceFrame *referenceFrame /* OPTIONAL */;
        GANSSTODm_t *ganssTODm /* OPTIONAL */;
        long *deltaGNASSTOD /* OPTIONAL */;
        GANSSTODUncertainty_t *ganssTODUncertainty /* OPTIONAL */;
        SeqOfGANSS_SgnTypeElement_t ganss_SgnTypeList;

        /* Context for parsing across buffer boundaries */
        asn_struct_ctx_t _asn_ctx;
    } GANSS_MsrSetElement_t;

    /* Implementation */
    extern asn_TYPE_descriptor_t asn_DEF_GANSS_MsrSetElement;

#ifdef __cplusplus
}
#endif

/* Referred external types */
#include "ReferenceFrame.h"

#endif /* _GANSS_MsrSetElement_H_ */
#include <asn_internal.h>
