/*
 * SPDX-FileCopyrightText: (c) 2003, 2004 Lev Walkin <vlm@lionet.info>. All rights reserved.
 * SPDX-License-Identifier: BSD-1-Clause
 * Generated by asn1c-0.9.22 (http://lionet.info/asn1c)
 * From ASN.1 module "ULP-Components"
 *     found in "../supl-common.asn"
 */

#include "MeasuredResults.h"

static asn_TYPE_member_t asn_MBR_MeasuredResults_1[] = {
    {ATF_POINTER, 3, offsetof(struct MeasuredResults, frequencyInfo),
        (ASN_TAG_CLASS_CONTEXT | (0 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_FrequencyInfo,
        0, /* Defer constraints checking to the member type */
        0, /* No PER visible constraints */
        0, "frequencyInfo"},
    {ATF_POINTER, 2, offsetof(struct MeasuredResults, utra_CarrierRSSI),
        (ASN_TAG_CLASS_CONTEXT | (1 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_UTRA_CarrierRSSI,
        0, /* Defer constraints checking to the member type */
        0, /* No PER visible constraints */
        0, "utra-CarrierRSSI"},
    {ATF_POINTER, 1, offsetof(struct MeasuredResults, cellMeasuredResultsList),
        (ASN_TAG_CLASS_CONTEXT | (2 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_CellMeasuredResultsList,
        0, /* Defer constraints checking to the member type */
        0, /* No PER visible constraints */
        0, "cellMeasuredResultsList"},
};
static int asn_MAP_MeasuredResults_oms_1[] = {0, 1, 2};
static ber_tlv_tag_t asn_DEF_MeasuredResults_tags_1[] = {
    (ASN_TAG_CLASS_UNIVERSAL | (16 << 2))};
static asn_TYPE_tag2member_t asn_MAP_MeasuredResults_tag2el_1[] = {
    {(ASN_TAG_CLASS_CONTEXT | (0 << 2)), 0, 0, 0}, /* frequencyInfo at 148 */
    {(ASN_TAG_CLASS_CONTEXT | (1 << 2)), 1, 0, 0}, /* utra-CarrierRSSI at 149 */
    {(ASN_TAG_CLASS_CONTEXT | (2 << 2)), 2, 0,
        0} /* cellMeasuredResultsList at 150 */
};
static asn_SEQUENCE_specifics_t asn_SPC_MeasuredResults_specs_1 = {
    sizeof(struct MeasuredResults),
    offsetof(struct MeasuredResults, _asn_ctx),
    asn_MAP_MeasuredResults_tag2el_1,
    3,                             /* Count of tags in the map */
    asn_MAP_MeasuredResults_oms_1, /* Optional members */
    3,
    0,  /* Root/Additions */
    -1, /* Start extensions */
    -1  /* Stop extensions */
};
asn_TYPE_descriptor_t asn_DEF_MeasuredResults = {
    "MeasuredResults",
    "MeasuredResults",
    SEQUENCE_free,
    SEQUENCE_print,
    SEQUENCE_constraint,
    SEQUENCE_decode_ber,
    SEQUENCE_encode_der,
    SEQUENCE_decode_xer,
    SEQUENCE_encode_xer,
    SEQUENCE_decode_uper,
    SEQUENCE_encode_uper,
    0, /* Use generic outmost tag fetcher */
    asn_DEF_MeasuredResults_tags_1,
    sizeof(asn_DEF_MeasuredResults_tags_1) /
        sizeof(asn_DEF_MeasuredResults_tags_1[0]), /* 1 */
    asn_DEF_MeasuredResults_tags_1,                /* Same as above */
    sizeof(asn_DEF_MeasuredResults_tags_1) /
        sizeof(asn_DEF_MeasuredResults_tags_1[0]), /* 1 */
    0,                                             /* No PER visible constraints */
    asn_MBR_MeasuredResults_1,
    3,                               /* Elements count */
    &asn_SPC_MeasuredResults_specs_1 /* Additional specs */
};
