/*
 * SPDX-FileCopyrightText: (c) 2003, 2004 Lev Walkin <vlm@lionet.info>. All rights reserved.
 * SPDX-License-Identifier: BSD-1-Clause
 * Generated by asn1c-0.9.22 (http://lionet.info/asn1c)
 * From ASN.1 module "SUPL-INIT"
 *     found in "../supl-init.asn"
 */

#ifndef _FormatIndicator_H
#define _FormatIndicator_H

#include <asn_application.h>

/* Including external dependencies */
#include <ENUMERATED.h>

#ifdef __cplusplus
extern "C"
{
#endif

    /* Dependencies */
    typedef enum FormatIndicator
    {
        FormatIndicator_logicalName = 0,
        FormatIndicator_e_mailAddress = 1,
        FormatIndicator_msisdn = 2,
        FormatIndicator_url = 3,
        FormatIndicator_sipUrl = 4,
        FormatIndicator_min = 5,
        FormatIndicator_mdn = 6,
        FormatIndicator_imsPublicIdentity = 7
        /*
         * Enumeration is extensible
         */
    } e_FormatIndicator;

    /* FormatIndicator */
    typedef ENUMERATED_t FormatIndicator_t;

    /* Implementation */
    extern asn_TYPE_descriptor_t asn_DEF_FormatIndicator;
    asn_struct_free_f FormatIndicator_free;
    asn_struct_print_f FormatIndicator_print;
    asn_constr_check_f FormatIndicator_constraint;
    ber_type_decoder_f FormatIndicator_decode_ber;
    der_type_encoder_f FormatIndicator_encode_der;
    xer_type_decoder_f FormatIndicator_decode_xer;
    xer_type_encoder_f FormatIndicator_encode_xer;
    per_type_decoder_f FormatIndicator_decode_uper;
    per_type_encoder_f FormatIndicator_encode_uper;

#ifdef __cplusplus
}
#endif

#endif /* _FormatIndicator_H_ */
#include <asn_internal.h>
