/*
 * SPDX-FileCopyrightText: (c) 2003, 2004 Lev Walkin <vlm@lionet.info>. All rights reserved.
 * SPDX-License-Identifier: BSD-1-Clause
 * Generated by asn1c-0.9.22 (http://lionet.info/asn1c)
 * From ASN.1 module "RRLP-Components"
 *     found in "../rrlp-components.asn"
 */

#include "GANSS-MsrSetElement.h"

static int memb_deltaGNASSTOD_constraint_1(
    asn_TYPE_descriptor_t *td, const void *sptr,
    asn_app_constraint_failed_f *ctfailcb, void *app_key)
{
    long value;

    if (!sptr)
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: value not given (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }

    value = *(const long *)sptr;

    if ((value >= 0 && value <= 127))
        {
            /* Constraint check succeeded */
            return 0;
        }
    else
        {
            _ASN_CTFAIL(app_key, td, sptr, "%s: constraint failed (%s:%d)",
                td->name, __FILE__, __LINE__);
            return -1;
        }
}

static asn_per_constraints_t ASN_PER_MEMB_DELTA_GNASSTOD_CONSTR_4 = {
    {APC_CONSTRAINED, 7, 7, 0, 127} /* (0..127) */,
    {APC_UNCONSTRAINED, -1, -1, 0, 0},
    0,
    0 /* No PER value map */
};
static asn_TYPE_member_t asn_MBR_GANSS_MsrSetElement_1[] = {
    {ATF_POINTER, 4, offsetof(struct GANSS_MsrSetElement, referenceFrame),
        (ASN_TAG_CLASS_CONTEXT | (0 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_ReferenceFrame,
        0, /* Defer constraints checking to the member type */
        0, /* No PER visible constraints */
        0, "referenceFrame"},
    {ATF_POINTER, 3, offsetof(struct GANSS_MsrSetElement, ganssTODm),
        (ASN_TAG_CLASS_CONTEXT | (1 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_GANSSTODm, 0,                  /* Defer constraints checking to the member type */
        0,                                      /* No PER visible constraints */
        0, "ganssTODm"},
    {ATF_POINTER, 2, offsetof(struct GANSS_MsrSetElement, deltaGNASSTOD),
        (ASN_TAG_CLASS_CONTEXT | (2 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_NativeInteger, memb_deltaGNASSTOD_constraint_1,
        &ASN_PER_MEMB_DELTA_GNASSTOD_CONSTR_4, 0, "deltaGNASSTOD"},
    {ATF_POINTER, 1, offsetof(struct GANSS_MsrSetElement, ganssTODUncertainty),
        (ASN_TAG_CLASS_CONTEXT | (3 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_GANSSTODUncertainty,
        0, /* Defer constraints checking to the member type */
        0, /* No PER visible constraints */
        0, "ganssTODUncertainty"},
    {ATF_NOFLAGS, 0, offsetof(struct GANSS_MsrSetElement, ganss_SgnTypeList),
        (ASN_TAG_CLASS_CONTEXT | (4 << 2)), -1, /* IMPLICIT tag at current level */
        &asn_DEF_SeqOfGANSS_SgnTypeElement,
        0, /* Defer constraints checking to the member type */
        0, /* No PER visible constraints */
        0, "ganss-SgnTypeList"},
};
static int asn_MAP_GANSS_MsrSetElement_oms_1[] = {0, 1, 2, 3};
static ber_tlv_tag_t asn_DEF_GANSS_MsrSetElement_tags_1[] = {
    (ASN_TAG_CLASS_UNIVERSAL | (16 << 2))};
static asn_TYPE_tag2member_t asn_MAP_GANSS_MsrSetElement_tag2el_1[] = {
    {(ASN_TAG_CLASS_CONTEXT | (0 << 2)), 0, 0, 0}, /* referenceFrame at 1456 */
    {(ASN_TAG_CLASS_CONTEXT | (1 << 2)), 1, 0, 0}, /* ganssTODm at 1457 */
    {(ASN_TAG_CLASS_CONTEXT | (2 << 2)), 2, 0, 0}, /* deltaGNASSTOD at 1458 */
    {(ASN_TAG_CLASS_CONTEXT | (3 << 2)), 3, 0,
        0}, /* ganssTODUncertainty at 1459 */
    {(ASN_TAG_CLASS_CONTEXT | (4 << 2)), 4, 0,
        0} /* ganss-SgnTypeList at 1463 */
};
static asn_SEQUENCE_specifics_t asn_SPC_GANSS_MsrSetElement_specs_1 = {
    sizeof(struct GANSS_MsrSetElement),
    offsetof(struct GANSS_MsrSetElement, _asn_ctx),
    asn_MAP_GANSS_MsrSetElement_tag2el_1,
    5,                                 /* Count of tags in the map */
    asn_MAP_GANSS_MsrSetElement_oms_1, /* Optional members */
    4,
    0,  /* Root/Additions */
    -1, /* Start extensions */
    -1  /* Stop extensions */
};
asn_TYPE_descriptor_t asn_DEF_GANSS_MsrSetElement = {
    "GANSS-MsrSetElement",
    "GANSS-MsrSetElement",
    SEQUENCE_free,
    SEQUENCE_print,
    SEQUENCE_constraint,
    SEQUENCE_decode_ber,
    SEQUENCE_encode_der,
    SEQUENCE_decode_xer,
    SEQUENCE_encode_xer,
    SEQUENCE_decode_uper,
    SEQUENCE_encode_uper,
    0, /* Use generic outmost tag fetcher */
    asn_DEF_GANSS_MsrSetElement_tags_1,
    sizeof(asn_DEF_GANSS_MsrSetElement_tags_1) /
        sizeof(asn_DEF_GANSS_MsrSetElement_tags_1[0]), /* 1 */
    asn_DEF_GANSS_MsrSetElement_tags_1,                /* Same as above */
    sizeof(asn_DEF_GANSS_MsrSetElement_tags_1) /
        sizeof(asn_DEF_GANSS_MsrSetElement_tags_1[0]), /* 1 */
    0,                                                 /* No PER visible constraints */
    asn_MBR_GANSS_MsrSetElement_1,
    5,                                   /* Elements count */
    &asn_SPC_GANSS_MsrSetElement_specs_1 /* Additional specs */
};
